{-
This covers the high level process of parsing. But not the parsing itself of the individual lines.

Process summary:
- make a stream tree
- parse all base level lines in parallel
- verify that the whole line was consumed
- apply same process recursively on bodies with syntax depending on header
-}
module Parsing.MainParsing where

import Parsing.Syntax
import Parsing.SrcLoc
import Parsing.LineParsing

import Common.Tree
import Common.Utils

import Process.Result

import Data.Map hiding (map)
import qualified List


--------------------------------------------------------------------------------
--	The output is a record of functions
--------------------------------------------------------------------------------

type TextParser  = String -> Result (Forest ParseTree) [ParseError]

data SyntaxParser = SyntaxParser {
	parseText :: TextParser, -- parses a whole text
	parseLine :: LineParser, -- parses starting rules and checks if the whole line was consumed
	parseRule :: String -> LineParser -- parses a specific rule
	}



toStreamForest :: String -> Forest Stream
toStreamForest = lines >>> indexed 1 >>> map toStream >>> List.filter (not . isBlank) >>> toForest isIndented where
	toStream (i,l) = (SrcLoc i 1, l)
	isBlank (_, str) = str `matches` "[ \t]*$"
	isIndented a b = countTabs b > countTabs a
	countTabs (i, str) = length (takeWhile (== '\t') str)




syntaxParser :: SyntaxGrammar -> SyntaxParser
syntaxParser grammar = SyntaxParser {
	parseText = _parseText,
	parseLine = _parseLine (startSyntaxes grammar),
	parseRule = _parseRule
	}
	where
		_parseRule :: String -> LineParser
		_parseRule = makeRuleParers grammar
		
		skip = makeSkip grammar

		allBodyRules :: Data.Map.Map String [String]
		allBodyRules = fromList (bodies grammar)

		
		_parseText :: TextParser
		_parseText = toStreamForest >>> parseBody (startSyntaxes grammar) >>> fSummarize
	
	
		parseBody :: [String] -> Forest Stream -> Forest (Result ParseTree ParseError)
		parseBody rules []     = []
		parseBody []    (t:ts) = [Tree (Failure (loc (content t), "Unexpected indented block.")) []]
		parseBody rules forest = map parseTree forest where
			parseTree :: Tree Stream -> Tree (Result ParseTree ParseError)
			parseTree tree = tree' where
				tree' = Tree {
					content = header,
					children = case header of
						Success h -> parseBody (bodyRules h) (children tree)
						Failure f -> []
					}
			
				header = _parseLine rules (content tree)
				
				
				--body rules forest = parseBody  (children tree)
				bodyRules (Tree root _) = case Data.Map.lookup (name root) allBodyRules of
					Just rules -> rules
					Nothing -> []
				

			
	
	
		_parseLine :: [String] -> LineParser
		_parseLine rules stream = _parseLine' (skip stream) where
			_parseLine' :: LineParser
			_parseLine' (loc, "")  = Success emptyNode
			_parseLine' (loc, str) = case parseEither rules defaultError of
				Success tree -> case skip ((loc, str) `after` tree) of
					(_, "")   -> Success tree
					(loc', _) -> Failure (loc', "End of line expected.")
				Failure f -> Failure f
				where
					rulesString = concat (List.intersperse ", " rules)
					
					defaultError :: ParseError
					defaultError = (loc, "Failed to parse either of " ++ rulesString ++ ". Unexpected characters: \"" ++ (startOf str) ++ "\"")
					
					parseEither :: [String] -> ParseError -> Result ParseTree ParseError
					parseEither [] err = Failure err
					parseEither (r:rs) err = case (_parseRule r) (loc, str) of
						Success s -> Success s
						Failure f -> parseEither rs (furthest err f)
		
